///////////////////////////////////////////////////////////////////////////////
//
//  MySSL.PRG
//
//  This is a common function used to create an xbSSLContext for a server-side
//  process.
//
//  Copyright (c) 2016, Xb2.NET inc.
//  All rights reserved.
//
//  Author:   Boris Borzic
//  Email:    support@xb2.net
//  Web:      http://www.xb2.net
//
///////////////////////////////////////////////////////////////////////////////

#include "xb2net.ch"

#define CERT_FILE         'MYCERT.PEM'      // SSL/TLS server certificate used for authentication
#define KEY_FILE          'MYCERT.PEM'      // We can combine the private key & certificate(s) in the same file
#define KEY_FILE_PWD      ''                // If the private key is password protected then specify the password here
#define DH_PARAMS_FILE    "DH2048.PEM"      // Load Diffie-Hellman parameters to be used for ephemeral keying
#define CIPHERLIST_FILE   "CipherList.txt"  // SSL/TLS cipher suites to be used by server
#define CRLF              (Chr(13)+Chr(10))

//-----------------------------------------------------------------------------
FUNCTION MySSLContext( lDestroy )
   Static oSSL
   Local cCipherList, cError

   if lDestroy == .T.
      if oSSL != NIL
         oSSL:destroy()
         oSSL := NIL
      endif
      Return Nil
   endif

   if oSSL == NIL
      oSSL := xbSSLContext():new( TLS_server_method )

      if !empty(KEY_FILE_PWD)
         oSSL:SetDefaultPassword(KEY_FILE_PWD)
      endif

      oSSL:SetOptions(SSL_OP_CIPHER_SERVER_PREFERENCE)   // use the server's cipher ordering preference, rather than the client's
      oSSL:SetOptions(SSL_OP_ALL)                        // enable workarounds for various bugs present in other SSL implementations
      oSSL:SetOptions(SSL_OP_NO_SSLv3)                   // do not use SSLv3 (comment out to support older browsers)
      oSSL:SetOptions(SSL_OP_SINGLE_DH_USE)              // do not reuse a DH key for different SSL sessions (improves forward secrecy at the expense of computational resources)
      oSSL:SetOptions(SSL_OP_SINGLE_ECDH_USE)            // do not reuse a ECDH key for different SSL sessions (improves forward secrecy at the expense of computational resources)
      oSSL:SetOptions(SSL_OP_NO_SESSION_RESUMPTION_ON_RENEGOTIATION) // disallow session resumption on renegotiation

      // No request for a certificate is sent. Any certificate received will be verified but will not terminate the handshake in a failure.
      oSSL:SetVerify( SSL_VERIFY_NONE )
      // uncomment next line if clients need a certificate in order to connect
      //oSSL:SetVerify(SSL_VERIFY_PEER + SSL_VERIFY_FAIL_IF_NO_PEER_CERT + SSL_VERIFY_CLIENT_ONCE)

      // load cipher list from file if available, otherwise use the default list
      cCipherList := MemoRead( CIPHERLIST_FILE )
      if empty(cCipherList)
         cCipherList := "AES256-GCM-SHA384:" // enable: AES in Galois Counter Mode (GCM): these ciphersuites are only supported in TLS v1.2.
         cCipherList += "AES128-GCM-SHA256:" // enable: AES in Galois Counter Mode (GCM): these ciphersuites are only supported in TLS v1.2.
         cCipherList += "!RC4-SHA:"          // disable: cipher suites using RC4. SSL3.0 + TLS 1.0
         cCipherList += "AES256-SHA256:"     // enable: cipher suites using 256 bit AES SSL3.0 + TLS 1.0 CBC (Cipher-block chaining)
         cCipherList += "AES128-SHA256:"     // enable: cipher suites using 128 bit AES SSL3.0 + TLS 1.0 CBC (Cipher-block chaining)
         cCipherList += "AES256-SHA:"        // enable: cipher suites using 256 bit AES SSL3.0 + TLS 1.0 CBC (Cipher-block chaining)
         cCipherList += "AES128-SHA:"        // enable: cipher suites using 128 bit AES SSL3.0 + TLS 1.0 CBC (Cipher-block chaining)
         cCipherList += "3DES:"              // enable: cipher suites using triple DES. SSL3.0 + TLS 1.0 (3DES is still used by WinXP/IE8)
         cCipherList += "HIGH:"              // enable: "high" encryption cipher suites (key lengths >= 128 bits)
         cCipherList += "MEDIUM:"            // enable: "medium" encryption cipher suites, currently some of those using 128 bit encryption.
         cCipherList += "!LOW:"              // disable: "low" encryption cipher suites, currently those using 64 or 56 bit encryption algorithms but excluding export cipher suites.
         cCipherList += "!aNULL:"            // disable: the cipher suites offering no authentication (these are vulnerable to a "man in the middle" attack)
         cCipherList += "!EXP:"              // disable: export encryption algorithms. Including 40 and 56 bits algorithms that can easily be broken.
         cCipherList += "!eNULL:"            // disable: all ciphers using no encryption
         cCipherList += "!ADH:"              // disable: Anonymous Diffie-Hellman (ADH) suites do not provide authentication.
         cCipherList += "!CAMELLIA:"         // disable: Camellia ciphersuites from RFC4132, extending TLS v1.0
         cCipherList += "!SEED:"             // disable: SEED ciphersuites from RFC4162, extending TLS v1.0
         cCipherList += "!MD5:"              // disable: cipher suites using MD5.
         cCipherList += "!DES:"              // disable: cipher suites using DES (not triple DES).
         cCipherList += "!IDEA"              // disable: cipher suites using IDEA.

         ErrorLog("CipherList file (" + CIPHERLIST_FILE + ") is empty, default= " + cCipherList)
      else
         cCipherList := strtran(cCipherList," ","")
         cCipherList := strtran(cCipherList,chr(10),"")
         cCipherList := strtran(cCipherList,chr(13),"")
         ErrorLog("CipherList= " + cCipherList)
      endif

      do case
      case !oSSL:SetCipherList(cCipherList)
         cError := "Problem with cipher list: " + cCipherList
      case ! oSSL:UseCertificateChainFile( CERT_FILE )
         cError := "Problem with certificate: " + CERT_FILE
      case ! oSSL:UsePrivateKeyFile( KEY_FILE )
         cError := "Problem with private key: " + KEY_FILE
      case ! oSSL:CheckPrivateKey()
         cError := "Private key does not match certificate public key!"
      case ! oSSL:LoadDHParams( DH_PARAMS_FILE )
         cError := "Problem loading DH params file: " + DH_PARAMS_FILE
      endcase

      if cError != NIL
         cError := "Unable to initialize SSL/TLS server - " + cError + CRLF + xbSSLGetErrorString()
         ErrorLog(cError, .T.)
         oSSL:destroy()
         oSSL := NIL
      endif

      // Generate 2048-bit temporary RSA key
      // The key length must be an even number with an upper limt of 4096
      // It is not a requirement for the RSA key to be set, but it does help to have it.
      oSSL:GenerateRSAKey( 2048 )
   endif

   Return oSSL

