///////////////////////////////////////////////////////////////////////////////
//
//  PROGRESSBAR.PRG
//
//  Demonstrates the use of "SendChunk" method to display a "real" progress bar
//  in a web browser. Adapted from original progressbar sample developed by
//  Alain Boucher, email: alain.boucher@acti-soft.com
//
//  Copyright (c) 2005-2015, Xb2.NET inc.
//  Portions Copyright (c) 2004, ACTI-SOFT
//  All rights reserved.
//
//  Author:   Boris Borzic
//  Email:    support@xb2.net
//  Web:      http://www.xb2.net
//
///////////////////////////////////////////////////////////////////////////////


#include "common.ch"
#include "thread.ch"

#xtranslate NTrim(<n>)  =>  LTrim(Str(<n>))

//-----------------------------------------------------------------------------
PROCEDURE WEB_ProgressBar()
   Local oBar

   oBar := HtmlProgressBar():new()

   oBar:HtmlHead    := '<title>Xb2.NET - Progress Bar</title>' +;
                       '<link rel="stylesheet" type="text/css" href="Style.css">' +;
                       '<meta name=viewport content="width=device-width, initial-scale=1, user-scalable=no">'
   oBar:HtmlMessage := '<b>This sample uses the "sendchunk" method to display a progressbar.</b><br><br><br>'

   oBar:Execute("TimeConsumingOperation")

   Return

//-----------------------------------------------------------------------------
CLASS HtmlProgressBar
   METHOD Draw

EXPORTED:
   VAR BarWidth, BarHeight, BarColor, BGColor, BorderColor
   VAR Cancelled        // signal to worker thread that operation is cancelled
   VAR ThreadPriority   // set priority of worker thread
   VAR HtmlHead         // custom HTML that will be inserted in <HEAD> section
   VAR HtmlMessage      // HTML formatted message to display while progress bar is active
   VAR Html             // the final HTML document generated by worker thread
   VAR CurrentStep      // current progress step, updated by worker thread
   VAR MaxSteps         // max number of steps in time consuming operation

   METHOD Execute

   INLINE METHOD Init()
      ::Cancelled      := .F.
      ::BarWidth       := "450"
      ::BarHeight      := "20"
      ::BarColor       := "blue"
      ::BGColor        := "white"
      ::BorderColor    := "blue"
      ::ThreadPriority := PRIORITY_NORMAL
      ::Html           := ""
      Return self

ENDCLASS

METHOD HtmlProgressBar:Draw()
   Local i, cHtml, cScript

   TEXT INTO cScript WRAP CHR(10) TRIMMED
      function ProgressBar(TableId,color1,color2,nStep) {
        var tbl  = document.getElementById(TableId);
        var rows = tbl.getElementsByTagName('tr');
        var cels = rows[0].getElementsByTagName('td');
        cels[0].innerHTML=nStep+"0%"
        for (var nCol=1; nCol<cels.length;nCol++) {
          if (nCol>nStep) {
              cels[nCol].style.background=color1;
          } else {
          	  cels[nCol].style.background=color2;
          }
        }
      }
   ENDTEXT

   cHtml := '<html><head>' + ::HtmlHead +;
            '<script language="JavaScript">' + cScript + '</script>' +;
            '</head>' +;
   	      '<body>' +;
            '<div id="ProgressBar" style="position:absolute;z-index:5;top:40%;left:15%;" align=center>' +;
            ::HtmlMessage

   // This is the actual progressbar
   cHtml += '<table id="progress" border=1 frame=border rules=none cellpadding=0 cellspacing=0 align=center '
   cHtml += 'BGcolor=' + ::BGColor + ' BorderColor=' + ::BorderColor
   cHtml += ' style="width:' + ::BarWidth + ';height:' + ::BarHeight + '; text-align:center;vertical-align:middle;">'
   cHtml += '<tr>'
   cHtml += '<td BGcolor=silver style="width:20%;height:100%;font:12px arial">0%</td>'
   for i := 1 to 10
      cHtml += '<td style="width:8%;height:100%;"></td>'
   next
   cHtml += '</tr>'
   cHtml += '</table>'

   Return cHtml


//-----------------------------------------------------------------------------
METHOD HtmlProgressBar:Execute( cbFunction )
   Local oThread
   Local cUpdateBar
   Local nStep, nLastStep, i
   Local oResponse  := ThreadObject():HTTPResponse


   // This piece of html will be sent to the browser to update the progressbar
   cUpdateBar := '<script language="JavaScript">ProgressBar("progress","' + ::BGColor + '","' + ::BarColor+'",'

   oResponse:TransferEncoding('chunked')
   oResponse:Send()
   oResponse:SendChunk( ::Draw() )

   ::CurrentStep := 0
   ::MaxSteps    := 10

   // start the time consuming operation in a separate thread
   oThread := Thread():New()
   oThread:SetPriority(::ThreadPriority)
   oThread:Start( cbFunction, self )

   // continue updating the progress bar while the time consuming operation is running
   while oThread:Active .and. !::Cancelled
      nStep := int( ::CurrentStep*10 / ::MaxSteps )

      if nStep != nLastStep .or. ++i >= 40
         nLastStep   := nStep
         ::Cancelled := (oResponse:SendChunk(cUpdateBar + NTrim(nStep) + ');</script>') < 0)

         // this ensures that we do a send every 12 seconds (40 x 300ms) even though the progress
         // step has not been updated by the worker thread. This way we can keep checking if the
         // user has closed the browser window or if the connection has been dropped
         i := 0
      end

      sleep(30)
   end

   if ! ::Cancelled
      oResponse:SendChunk('</div><script language="JavaScript">document.getElementById("ProgressBar").style.display="none";</script>')
      oResponse:SendChunk(::Html)
      oResponse:SendChunk( "</body></html>" )
      oResponse:SendChunk()
   endif

   Return


//-----------------------------------------------------------------------------
PROCEDURE TimeConsumingOperation( oProgress )
   Local nStarted := Seconds()
   Local i := 0
   Local cHtml := "<h1>This is my time consuming report...</h1><p>"

   // Run for 15 seconds or until operation is cancelled by user.
   // For reasons of simplicity, we simply update the progress step every second.
   // In a real app you would probably set :MaxSteps to the total # of major steps
   // in the time consuming operation, and then update :CurrentStep at the end
   // of each step.

   oProgress:MaxSteps := 15

   // keep checking if user has cancelled this process
   while ! oProgress:Cancelled

      if Seconds()-nStarted > oProgress:MaxSteps
         oProgress:CurrentStep := oProgress:MaxSteps
         exit
      end

      oProgress:CurrentStep := Seconds()-nStarted
      sleep(30)
      cHtml += "<br>Line " + NTrim(++i)

   end

   oProgress:Html := cHtml

   Return


