///////////////////////////////////////////////////////////////////////////////
//
//  TCLIENT.PRG
//
//  Xb2.NET sample showing how to create a simple telnet client
//
//  Copyright (c) 2001-2016, Xb2.NET inc.
//  All rights reserved.
//
//  Author:   Boris Borzic
//  Email:    support@xb2.net
//  Web:      http://www.xb2.net
//
///////////////////////////////////////////////////////////////////////////////

#include "xb2net.ch"
#include "common.ch"
#include "xbp.ch"
#include "inkey.ch"

#pragma  library ("xb2net.lib")

#define CRLF           (Chr(13)+Chr(10))
#xtranslate NTrim(<n>) =>  LTrim(Str(<n>))

//-----------------------------------------------------------------------------
static procedure ParamHelp()
  ?
  ? "Xb2.NET sample telnet client"
  ?
  ? "Parameters:"
  ? "  /HOST:string  - Telnet server host name or IP address"
  ? "  /PORT:number  - Port number of remote Telnet server"
  ? "  /SSL          - Create secure connection when specified"
  ?
  ? "Example:"
  ? "  TCLIENT /HOST:127.0.0.1 /PORT:992 /SSL"
  ?
  inkey(0)
  quit
  Return

//-----------------------------------------------------------------------------
procedure Main()
   Local i, nKey, cSend, cRecv, oSock, cColor, cHost, cPort, lSecure, oSSLContext
   Local cParamLine := ""

   for i := 1 to pcount()
      cParamLine += pvalue(i) + " "
   next

   if "/?" $ cParamLine .or. "/HELP" $ cParamLine
      ParamHelp()
   endif

   cHost  := GetParam("/HOST:", cParamLine)
   cPort  := GetParam("/PORT:", cParamLine)
   lSecure:= GetParam("/SSL"  , cParamLine) != NIL

   DEFAULT cHost TO "LocalHost"              // a server running on the same machine
   DEFAULT cPort TO iif(lSecure,"992","23")  // default port for Telnet

   SET CURSOR ON

   if lSecure .and. (oSSLContext := GetSSLContext()) == NIL
      inkey(5)
      Return
   endif

   ? "create socket..."
   oSock := xbSocket():New(AF_INET, SOCK_STREAM, IPPROTO_IP, oSSLContext)

   if oSock:ErrorCode > 0
      DispError("Unable to create a socket!", oSock, oSSLContext)
      Return
   endif

   ? "connect to server..."

   if ! oSock:Connect(cHost, Val(cPort))
      DispError("Unable to connect to: " + cHost, oSock, oSSLContext)
      Return
   endif

   // turn off blocking mode so that :Send/:Recv don't wait
   oSock:SetBlockingMode(.f.)

   ? "connected to: ", oSock:RemoteName, ", address:", oSock:InetNtoA(oSock:RemoteAddr), ", port:", NTrim(oSock:RemotePort)

   nKey := 1

   while nKey != K_ESC
      if nKey != 0
         if oSock:Send( cSend ) < 0
            ? "connection lost..."
            exit
         endif
      endif

      cRecv := oSock:Recv()

      do case
      case cRecv == NIL
         ? "connection lost..."
         exit

      case Empty(cRecv)
         nKey := inkey(0.1)
         if nKey > 0
            cSend := iif(nKey == K_ENTER, CRLF, chr(nKey))
            ?? cSend
         endif

      otherwise
         SET CURSOR OFF
         nKey   := 0
         cColor := SetColor("w/b")
         ? cRecv
         SetColor(cColor)
         ?
         SET CURSOR ON

      endcase
   end

   FreeResources( oSock, oSSLContext )
   wait
   Return

//-----------------------------------------------------------------------------
static procedure DispError(cMsg, oSock, oSSLContext)
   MsgBox(cMsg + chr(10) + "Error code: " + NTrim(oSock:ErrorCode) + " (" + oSock:ErrorMessage + ")")
   FreeResources( oSock, oSSLContext )
   Return

//-----------------------------------------------------------------------------
static procedure FreeResources( oSock, oSSLContext )
   oSock:destroy()
   if oSSLContext != NIL
      oSSLContext:destroy()
   endif
   Return

//-----------------------------------------------------------------------------
static function GetParam( cTag, cString )
   Local i, cVal

   cTag := upper(cTag)

   if (i := at(upper(cTag), upper(cString))) == 0
      Return nil
   endif

   cVal := SubStr(cString, i+len(cTag))

   if (i := at(" ", cVal)) == 0
      Return cVal
   endif
   Return left(cVal, i-1)

//-----------------------------------------------------------------------------
static function GetSSLContext()
   Local oError, bErrBlock, cColor, oSSLContext

   // Possible reasons why the OpenSSL library cannot be loaded include:
   // 1. Your Xb2NetKey is invalid or it does not include SSL support
   // 2. The OpenSSL LIB files: SSLEAY32.DLL and LIBEAY32.DLL cannot be found in the system path

   bErrBlock := ErrorBlock( {|x|break(x)})
   BEGIN SEQUENCE
      oSSLContext := xbSSLContext():new( TLS_client_method )
      oSSLContext:SetOptions(SSL_OP_NO_SSLv3)
      ? "Using", oSSLContext:Version[SSLEAY_VERSION]
      ?

   RECOVER USING oError
      cColor := SetColor("r+/n")
      ?
      ? "ERROR: Unable to create SSL context!"
      ? oError:Description
      ? "Do you have the OpenSSL library installed?"
      ?
      SetColor(cColor)

   ENDSEQUENCE
   ErrorBlock( bErrBlock )

   Return oSSLContext


