///////////////////////////////////////////////////////////////////////////////
//
//  TSERVER.PRG
//
//  Xb2.NET sample showing how to create a simple Telnet server
//
//  Copyright (c) 2001-2016, Xb2.NET inc.
//  All rights reserved.
//
//  Author:   Boris Borzic
//  Email:    support@xb2.net
//  Web:      http://www.xb2.net
//
///////////////////////////////////////////////////////////////////////////////


#include "xb2net.ch"
#include "common.ch"
#include "xbp.ch"
#include "error.ch"
#include "inkey.ch"
#include "thread.ch"

#pragma  library ("xb2net.lib")

#define APP_NAME        'Xb2.NET sample Telnet server'
#define DEF_PORT        23   // the standard Telnet port
#define DEF_SPORT       992  // the standard port for Telnet over SSL/TLS
#define CRLF            (Chr(13)+Chr(10))

#xtranslate NTrim(<n>)  =>  LTrim(Str(<n>))


//-----------------------------------------------------------------------------
procedure Main()
   Local i, aThread, cTitle, bErrBlock, oError, cColor, oServer, oSServer
   Local nCols := MaxCol() + 1

   SetPos(MaxRow(),0)

   if (oServer := TelnetServer( DEF_PORT )) == NIL
      Terminate(5)
   endif

   // check if the OpenSSL library can be loaded
   //
   // Possible reasons why the OpenSSL library cannot be loaded include:
   // 1. Your Xb2NetKey is invalid or it does not include SSL support
   // 2. The OpenSSL LIB files: SSLEAY32.DLL and LIBEAY32.DLL cannot be found in the system path

   bErrBlock := ErrorBlock( {|x|break(x)})
   BEGIN SEQUENCE
      if xbSSLAvailable()
         // start secure Telnet server in it's own thread
         oSServer := TelnetServer( DEF_SPORT, .t. )
      endif

   RECOVER USING oError
      ErrorLog("ERROR: Secure Telnet server cannot be started!" + CRLF +;
               oError:Description + CRLF +;
               "Do you have the OpenSSL library installed?", .T.)

   ENDSEQUENCE
   ErrorBlock( bErrBlock )


   cTitle  := PadM(APP_NAME + " " + xbSocket():Version, "(ESC)=Exit", nCols)

   while .t.
      DispOutAt(0,0, cTitle, "gr+/b" )
      DispOutAt(1,0, PadR(GetStatus(oServer, oSServer), nCols), "gr+/b" )
      DispOutAt(2,0, Space(nCols), "w/n")

      aThread := ThreadInfo(THREADINFO_TOBJ)
      for i := Len(aThread) to 1 step -1
         if aThread[i,1] != NIL .and. aThread[i,1]:Active
            DispOutAt(2, i-1, "*", "g/n")
         endif
      next

      if inkey(1) == K_ESC
         if ConfirmBox(nil, "Terminate " + APP_NAME + "?", "Quit", ;
            XBPMB_OKCANCEL, XBPMB_WARNING+XBPMB_MOVEABLE+XBPMB_APPMODAL, XBPMB_DEFBUTTON2) == XBPMB_RET_OK
            exit
         endif
      endif
   end

   Terminate(5, oServer, oSServer)

   Return

//-----------------------------------------------------------------------------
PROCEDURE Terminate( nPause, oServer, oSServer )
   ErrorLog("Terminating Telnet server...")

   if oServer != NIL
      oServer:onError := NIL
      oServer:Destroy()  // don't forget to destroy the object in order to release socket handles!
   endif
   if oSServer != NIL
      oSServer:onError := NIL
      oSServer:Destroy()  // don't forget to destroy the object in order to release socket handles!
      MySSLContext(.t.)   // also destroy the SSLContext since we don't need it any longer
   endif
   inkey(nPause)
   Quit
   Return

//-----------------------------------------------------------------------------
PROCEDURE ErrorLog( cMsg, lHilite )
   Local cColor := iif(lHilite==.T., SetColor("r+/n"), SetColor("bg/n"))
   ?
   ? cMsg                            // display on screen, and
   SetColor(cColor)
   //xbErrorLog(cMsg,,, ERR_LOG_FILE)  // log to file
   Return

//-----------------------------------------------------------------------------
STATIC FUNCTION GetStatus( oServer, oSServer )
   Local n1 := oServer:ActiveConnections()
   Local n2 := oServer:ConnectCount
   Local n3 := Len(oServer:ThreadPool)

   if oSServer != NIL
      n1 += oSServer:ActiveConnections()
      n2 += oSServer:ConnectCount
      n3 += Len(oSServer:ThreadPool)
   endif

   Return "ActiveConnections:" + Str(n1,4) +;
          "; ThreadPool:" + Str(n3,4) +;
          ";  TotalHits: " + Str(n2,7)

//-----------------------------------------------------------------------------
FUNCTION PadM( cStr1, cStr2, nLen, cFillChar )
   Return Left(cStr1 + PadL(cStr2, nLen - Len(cStr1), cFillChar), nLen)

//-----------------------------------------------------------------------------
STATIC FUNCTION TelnetServer( nPort, lSecure )
   Local i, oSSL, oServer

   DEFAULT lSecure TO .F.

   if lSecure
      if (oSSL := MySSLContext()) == NIL
         Return NIL
      endif
   endif

   oServer := xbServer():new(INADDR_ANY, nPort, oSSL, "TelnetServer")

   if oServer:ErrorCode > 0
      ErrorLog("Unable to start server on port: " + NTrim(nPort) + CRLF +;
               "Error code: " + NTrim(oServer:ErrorCode) + " (" + oServer:ErrorMessage + ")", .T.)
      oServer:destroy()
      MySSLContext(.t.) // destroy the SSLContext since we can't start the server
      Return NIL
   endif

   oServer:LingerTimeout(60)                // wait up to 60 seconds for buffered data to be received by client
   oServer:onConnect := {||ServiceClient()} // this is the codeblock that will be evaluated when a client connection is accepted
   oServer:atEnd     := {||DbCloseAll()}    // ensure all database files are closed when client connection is terminated
   oServer:RecvTimeout(1000)

   oServer:start()  // start the Telnet server in it's own thread

   ? "Telnet server started on host:", oServer:LocalName, "; Port:", NTrim(nPort)
   if lSecure
      ? "Using", oSSL:Version[SSLEAY_VERSION]
   endif

   Return oServer

//-----------------------------------------------------------------------------
procedure ServiceClient()
   Local cRecv, cSend, cFileName, cColor, oError
   Local cThreadID := Str(ThreadID(),3)
   Local oClient   := ThreadObject()

   ErrorBlock({|e|Break(e)})

   BEGIN SEQUENCE
      cColor := SetColor("w+/n")
      ? cThreadID, "** CONNECT: RemoteName:", oClient:RemoteName, " Address:", oClient:InetNtoA(oClient:RemoteAddr), ;
         "Port:", NTrim(oClient:RemotePort), iif(oClient:isSecure, "via SSL", "")
      SetColor(cColor)

      // send welcome message to client with some instructions
      oClient:Send( "Hello " + oClient:RemoteName + CRLF +;
                    "You are connected to my Xb2.NET Telnet server" + CRLF +;
                    "Try typing a DOS command like DIR..." + CRLF )

      // temp file name to redirect output of DOS commands to
      cFileName := "~" + NTrim(oClient:Handle) + ".TMP"

      // do not timeout on receive
      oClient:RecvTimeout(0)

      // wait until we receive something on client socket (or error occurs)
      // remember this is running in it's own thread

      while (cRecv := oClient:RecvLine()) != NIL
         cSend := cRecv
         cRecv := StrTran(StrTran(cRecv,chr(10),""),chr(13),"")
         ? cThreadID, cRecv

         // execute DOS command received from client
         // redirecting output to a temp file
         RunShell( "/C " + cRecv + " > " + cFileName,,.f.,.t. )

         // read contents from temp file so we can send it to the client
         cSend += MemoRead(cFileName)
         FErase(cFileName)

         if oClient:Send( cSend ) < 0
            exit
         endif
      end

      cColor := SetColor("w/r")
      ? cThreadID, "** QUIT **"
      SetColor(cColor)

   RECOVER USING oError
      cColor := SetColor("w/r")
      ? cThreadID, "** ERROR **", oError:description, oError:operation, oError:args, oError:filename
      SetColor(cColor)

   END SEQUENCE

   oClient:destroy()
   Return


