///////////////////////////////////////////////////////////////////////////////
//
//  WSCLIENT.PRG
//
//  Xb2.NET sample showing how to create a simple WebSocket client
//
//  Copyright (c) 2015, Xb2.NET inc.
//  All rights reserved.
//
//  Author:   Boris Borzic
//  Email:    support@xb2.net
//  Web:      http://www.xb2.net
//
//  You can specify a URL on the command line as follows:
//
//  WSCLIENT wss://echo.websocket.org
//  WSCLIENT ws://localhost/echo
//  WSCLIENT ws://localhost/stream
//  WSCLIENT ws://127.0.0.1/echo
//  WSCLIENT ws://127.0.0.1/stream
//
//  Note: To establish a secure WebSocket connection use the 'wss://' prefix, eg:
//
//  wss://echo.websocket.org   (this connection is secure, using SSL/TLS)
//  ws://echo.websocket.org    (this connection is not secure)
//
///////////////////////////////////////////////////////////////////////////////


#include "xb2net.ch"
#include "common.ch"
#include "xbp.ch"
#include "inkey.ch"

#pragma  library ("xb2net.lib")

#define DISCONNECT  1
#define CONNECTING  2
#define CONNECT     3
#define SEND        4
#define RECV        5
#define PING        6
#define PONG        7

//-----------------------------------------------------------------------------
procedure Main( cURL )
   Local oSock, nKey, oThread, cSend := ""

   DEFAULT cURL TO "ws://xb2.net/echo"

   // make sure that the URL is an absolute reference
   cURL := lower(alltrim(cURL))
   if ! left(cURL, 5) == "ws://"  .and. ! left(cURL, 5) == "ws:\\" .and. ;
      ! left(cURL, 6) == "wss://" .and. ! left(cURL, 6) == "wss:\\"
      cURL := "ws://" + cURL
   endif

   // make the console window bigger
   SetAppWindow():Configure( NIL, NIL, {0,23}, 35, 95, AppName() + " " + cURL)

   oSock := xbWebSocket():new()
   StatusLog( CONNECTING, cURL )

   ?? "connecting to"
   ? cURL

   if !oSock:Connect(cURL)
      MsgBox("Unable to connect to: " + cURL + chr(10) +;
             "Error code: " + LTrim(Str(oSock:ErrorCode)) + " (" + oSock:ErrorMessage + ")" )
      Return
   endif

   StatusLog( CONNECT )

   ? "connected"
   if oSock:isSecure()
      ?? " (using secure connection)"
      setcolor("w+/n")
      ? oSock:GetCurrentCipher(.t.)
      ?
   endif

   setcolor("g/n")
   ? "Type a message, hit ENTER to send"
   ? "ESC to exit..."
   ?
   ?
   setcolor("w/n")

   // disable blocking mode so that :recv() will not block
   oSock:SetBlockingMode(.f.)

   // start a new thread to listen for messages coming from server
   oThread := Thread():new()
   oThread:start("RECEIVE", oSock)

   SET CURSOR ON

   while !oSock:Terminated
      nKey := inkey(0.1)

      if nKey > 0
         if nKey == K_ESC
            oSock:close() // send disconnect message to server
            exit
         elseif nKey == K_ENTER
            StatusLog( SEND, cSend )
            if oSock:send(cSend) == SOCKET_ERROR
               StatusLog( DISCONNECT, oSock:ErrorMessage )
               exit
            endif
            ?
            cSend := ""
         else
            ?? chr(nKey)
            cSend += chr(nKey)
         endif
      endif
   end

   setcolor("r+/n")
   ?
   ? "disconnected"
   oSock:destroy()
   xbSSLFreeErrorQueue()

   sleep(100)
   wait
   Return

//-----------------------------------------------------------------------------
procedure Receive( oSock )
   Local cRecv, cError

   // continue listening for messages until disconnected

   while !oSock:Terminated
      cRecv := oSock:Recv()
      if cRecv == NIL
         // send a PING and wait up to 4 seconds for server to reply
         // if a PONG is not received in that time, then terminate connection
         ? "send ping", oSock:ErrorCode, oSock:ErrorMessage
         if !oSock:Ping("", 4000, @cError)
            ? "exit PING failed"
            exit
         endif
      elseif oSock:OpCode == NIL
         // there is no incoming data available
         sleep(20)
      elseif oSock:OpCode == WS_OP_PING
         StatusLog( PING, cRecv )
      elseif oSock:OpCode == WS_OP_PONG
         StatusLog( PONG, cRecv )
      else
         StatusLog( RECV, cRecv )
      endif
   end

   ? "exiting receive thread"

   oSock:Terminated := .T. // signal to main thread
   StatusLog( DISCONNECT, cError )
   Return

//-----------------------------------------------------------------------------
procedure StatusLog( nStatus, cMessage )
   Static oXbp, oWindow
   Static aColor  := {"r+/w","rb/w", "g/w","b/w","w/b","bg/w","bg/w"}
   Static aStatus := {"DISCONNECTED","CONNECTING","CONNECTED","SEND:","RECV:","PING:","PONG:"}

   if oXbp == nil
      oXbp := XbpCrt():new(,,{354,110},30,45,"LOG")
      oXbp:autoFocus := .F.
      oXbp:create()
      oWindow := SetAppWindow(oXbp)
      SetColor("n/w")
      cls
   endif

   SetAppWindow(oXbp)
   SetColor(aColor[nStatus])

   if cMessage == NIL .or. cMessage == ""
      ? aStatus[nStatus]
   else
      ? aStatus[nStatus], cMessage
   endif

   SetAppWindow(oWindow)
   Return



