///////////////////////////////////////////////////////////////////////////////
//
//  WEBFUNC.PRG
//
//  Copyright (c) 2002-2016, Xb2.NET inc.
//  All rights reserved.
//
//  Author:   Boris Borzic
//  Email:    support@xb2.net
//  Web:      http://www.xb2.net
//
//
//  ** NOTE **
//  The default 'onGET' & 'onPOST' HTTP event handlers will trigger an Xbase++
//  function with a name analogous to the HTTP request path and a prefix of
//  'WEB_' or 'SWEB_'. Backslash (\) characters in the path are replaced with
//  an underscore character (_). An optional list of parameters may also be
//  passed directly to the Xbase++ function when enclosed within parentheses.
//  Example:
//
//  URI:                                      Xbase++ function:
//  ---------------------------------------   ---------------------------
//  http://localhost/login?                   WEB_login()
//  http://localhost/tc/login?                WEB_tc_login()
//  https://mysite.com/GetImage('123',.T.)?   SWEB_GetImage('123',.T.)
//  https://mysite.com/pdf/view('AP1402')?    SWEB_pdf_view('AP1402')
//
//  A secure connection initiated with the 'https' prefix can call a function with
//  either the 'WEB_' or 'SWEB_' prefix. However, a non-secure connection can only
//  call a 'WEB_' function.
//
//
//  The following functions are exported and can be called by an HTTP web client:
//
//    WEB_Run( cAppName )
//    WEB_Exec( cCommand )
//    WEB_Eval( xCode )
//    WEB_View( cFileName )
//    WEB_ViewSource( cFileName )
//    WEB_DirectFunctionCall()
//    WEB_CRC32( cFileName )
//    WEB_SendChunks()
//    WEB_Upload()
//    WEB_WWWAuthenticate()
//    WEB_Login()
//    WEB_Logout()
//    WEB_ChangePassword()
//    WEB_BrowseTable()
//    WEB_AjaxTest( nAction )
//    WEB_CometTest( nAction )
//
///////////////////////////////////////////////////////////////////////////////


#include "xb2net.ch"
#include "common.ch"
#include "fileio.ch"

#pragma  library ("xb2net.lib")

#define CRLF                (Chr(13)+Chr(10))
#define VIEWPORT            '<head><meta name=viewport content="width=device-width, initial-scale=1"></head>'

#xtranslate NTrim(<n>)  =>  LTrim(Str(<n>))


//-----------------------------------------------------------------------------
// Execute an application on the web server.
//
PROCEDURE WEB_Run( cAppName )
   Local oClient := ThreadObject()

   DEFAULT cAppName TO oClient:GetVar(1)

   if oClient:LocalAddr != oClient:RemoteAddr
      // prevent client from running an app if they are connecting through a remote machine
      FlashError( '** ERROR **<p>Application: <i>' + cAppName + ;
                  '</i> can only be executed from a browser running on the same machine as the web server!',,8,.t.)
   else
      cAppName := StrTran(StrTran(cAppName,'"',""),"'",'') // remove any quotes around value
      RunShell( "/C START /B " + cAppName,,.t., .f. )
      FlashMessage('Application: <b>' + cAppName +'</b> has been executed.')
   endif

   Return

//-----------------------------------------------------------------------------
// Execute a DOS command (only DIR and CHKDSK commands are allowed)
//
PROCEDURE WEB_Exec( cCommand )
   Local oClient := ThreadObject()
   Local cTempFile

   DEFAULT cCommand TO oClient:GetVar(1)

   if Empty(cCommand)
      FlashError('Nothing to execute...')
      Return
   endif

   // create a unique temp file name to redirect output of DOS commands to
   cTempFile := "~" + NTrim(oClient:Handle) + ".TMP"
   cCommand  := Upper(Trim(cCommand))

   if Left(cCommand,3) == "DIR" .or. Left(cCommand,6) == "CHKDSK"
      RunShell( "/C " + cCommand + " > " + cTempFile,,.f.,.t. )
      // read contents from temp file so we can send it to the client
      oClient:HTTPResponse:ContentType := "text/plain"
      oClient:HTTPResponse:Content     := MemoRead(cTempFile)
      FErase(cTempFile)
      if Empty(oClient:HTTPResponse:Content)
         oClient:HTTPResponse:Content := cCommand
      endif
   else
      FlashError('** ERROR **<p>Command <i>' + cCommand +'</i> is not allowed!',,,.t.)
   endif

   Return

//-----------------------------------------------------------------------------
// Evaluate a codeblock or macro expression.
// NOTE: This can be a little dangerous, it is only provided as an example of what can be done.
//
PROCEDURE WEB_Eval( xCode )
   Local oClient := ThreadObject()
   Local xValue

   DEFAULT xCode TO  oClient:GetVar(1)

   if Empty(xCode)
      FlashError('Nothing to evaluate...')
      Return
   endif

   if ValType(xCode) == "C"
      xValue := &(xCode)
   else
      xValue := xCode
   endif

   if ValType(xValue) == "B"
      xValue := Eval(xValue)
   endif

   oClient:HTTPResponse:Content := VIEWPORT +;
      '<html><body><code>Results of EVAL( <b>'+ Var2Char(xCode) + '</b> )<p>' +;
      Var2Char(xValue) +;
      '</code></body></html>'

  Return

//-----------------------------------------------------------------------------
// Send back contents of requested file as "text/plain"
//
// NOTE: The file location is relative to the directory where the server app is running,
// which may not be the same as the www root directory specified in xbHTTPServer:RootDir
//
PROCEDURE WEB_View( cFileName )
   Local oClient := ThreadObject()

   DEFAULT cFileName TO oClient:GetVar(1)

   if Empty(cFileName) .or. "../" $ cFileName .or. "..\" $ cFileName .or. ":" $ cFileName .or. cFileName[1] $ "/\"
      FlashError('Invalid file name...')
      Return
   endif

   cFileName := StrTran(StrTran(cFileName,'"',""),"'",'') // remove any quotes around value

   if FExists( cFileName )
      oClient:HTTPResponse:ContentType := "text/plain"
      oClient:HTTPResponse:Content     := MemoRead(cFileName)
   else
      FlashError('** ERROR **<p>File: <i>'+ cFileName + '</i> is not found in server directory!',,,.t.)
   endif

   Return

//-----------------------------------------------------------------------------
// This function is used to send a file to the client. It is used to bypass
// execution of .PRG files so that the file contents can be sent to client.
// The file path is relative to the www root folder.
//
PROCEDURE WEB_ViewSource( cPath )
   Local oClient := ThreadObject()

   DEFAULT cPath TO oClient:GetVar(1)

   if Empty(cPath) .or. "../" $ cPath .or. "..\" $ cPath .or. ":" $ cPath
      oClient:NotFound()
      Return
   endif

   oClient:SendFile( oClient:Parent:RootDir + iif(cPath[1] $ "/\", SubStr(cPath,2), cPath) )
   Return

//-----------------------------------------------------------------------------
PROCEDURE WEB_DirectFunctionCall()
   Static cTemplate
   Local i, cRet, cParams := "", imax := PCount()

   if cTemplate == NIL
      TEXT INTO cTemplate TRIMMED
      <html>
      <head>
        <title>Passing parameters directly to Xbase++ function</title>
        <meta name=viewport content="width=device-width, initial-scale=1">
        <link rel="stylesheet" type="text/css" href="Style.css">
      </head>
      <body bgcolor=black style=margin:0px;>
        <table width=90% align=center>
          <tr><th colspan=3>WEB_DirectFunctionCall()</th></tr>
          <tr><th>Param #</th><th>ValType</th><th>Value</th></tr>
          %params%
          <tr>
          <td colspan=3>
            <b>This is the HTML <u>client-side</u> code:</b><p>
            <code>
               &lt;a href=%uri%&gt;Direct function call&lt;/a&gt;
            </code>
            <br><br><br>
            NOTE: Data within the HREF tag must be URL encoded. Use the <b>xbEscape() </b>
            function to automatically encode URL strings generated by your Xb2.NET HTTP server application.
          </td>
          </tr>
        </table>
      </body>
      </html>
      ENDTEXT
   endif

   for i := 1 to imax
      cParams += "<tr><td>" + NTrim(i) + "</td><td>" +;
                 ValType(PValue(i)) + "</td><td>" +;
                 Var2Char(PValue(i)) + "</td><tr>"
   next

   cRet := StrTran(cTemplate, "%uri%", ThreadObject():HTTPRequest:xbURI:AsString())
   cRet := StrTran(cRet, "%params%", cParams)

   ThreadObject():HTTPResponse:Content := cRet
   Return

//-----------------------------------------------------------------------------
PROCEDURE WEB_CRC32( cFileName )
   Local cBuffer, nFile, nBytes, nCRC := 0

   if Empty(cFileName)
      FlashError('No file specified...')
      Return
   endif

   nFile := FOpen( cFileName, FO_READ )

   if nFile < 0
      FlashError('** ERROR **<p>Unable to open file: <i>'+ cFileName + '</i><p>('+ DosErrorMessage(FError()) + ')',,,.t.)
   endif

   cBuffer := Space(32768)

   while (nBytes := FRead(nFile, @cBuffer, 32768)) > 0
      nCrc := xbZCRC32(nCrc, cBuffer, nBytes)
   end

   FClose(nFile)

   ThreadObject():HTTPResponse:Content := VIEWPORT +;
      '<html><body><code>CRC32 for file: ' + cFileName +;
      ' = <b>' + NTrim(nCRC) + '</b></code></body></html>'

   Return


//-----------------------------------------------------------------------------
PROCEDURE WEB_SendChunks()
   Local i, cHtml, oResponse

   // get reference to current xbHTTPResponse object
   // ThreadObject() will return a reference to the current xbHTTPThread object

   oResponse := ThreadObject():HTTPResponse

   TEXT INTO cHtml TRIMMED
      <html><head>
      <title>Xb2.NET - Sending Chunks</title>
      <meta name=viewport content="width=device-width, initial-scale=1">
      <link rel="stylesheet" type="text/css" href="Style.css"></head>
      <body>
      <b>
      This sample shows how to send dynamic content in discrete chunks.<br>
      The server will pause for one second after sending each chunk of data:
      </b>
      <br><br><ul>
   ENDTEXT

   // add a transfer-encoding header field and set it to "chunked"
   oResponse:TransferEncoding('chunked')

   // send the message header (the content will be sent later)
   oResponse:Send()

   // send first chunk
   oResponse:SendChunk(cHtml)

   for i := 1 to 10
      if oResponse:SendChunk('This is chunk # ' + NTrim(i) + '<br>') == SOCKET_ERROR
         // connection terminated
         Return
      endif
      // pause for one second before sending the next chunk
      sleep(100)
   next

   oResponse:SendChunk('</ul><br><hr>All done</body></html>')

   // send an empty chunk to signal end of content
   oResponse:SendChunk()

   Return

//-----------------------------------------------------------------------------
// process MIME or DIME attachments uploaded to HTTP server
//
PROCEDURE WEB_Upload()
   Local i, aPayload, cFileName, cUserName, lSuccess
   Local oClient    := ThreadObject()
   Local oResponse  := oClient:HTTPResponse
   Local oPayload   := oClient:HTTPRequest:Content
   Local cUploadDir := oClient:Parent:RootDir + "upload\"

   oResponse:ContentType:= "text/plain"

   if ValType(oPayload) != "O" .or. ! oPayload:isDerivedFrom("xbPayload")
      oResponse:StatusCode := 406
      oResponse:Content    := "The server expected a MIME or DIME payload."
      Return
   endif

   // make sure upload folder exists
   CreateDir(cUploadDir)

   // we are expecting a "UserName" form var to be passed
   cUserName := oClient:GetVar("UserName")

   if Empty(cUserName)
      oResponse:Content := ""
   else
      oResponse:Content := "Thank you " + cUserName + "!" + CRLF
      // attachments will be saved in the .\upload\user_name folder
      cUploadDir += CleanFileName(cUserName) + "\"
      CreateDir(cUploadDir)
   endif

   // get the payload list
   aPayload := oPayload:List

   oResponse:Content += "The following attachments have been saved in the server's UPLOAD folder:" +CRLF+CRLF

   // save the payloads in the upload folder
   for i := 1 to len(aPayload)
      if empty(aPayload[i,PL_DATA])
         // skip empty payloads
         loop
      endif

      cFileName := xbXMLGetAttribute(aPayload[i,PL_OPTIONS], "filename")

      if empty(cFileName)
         // if a filename is not provided, use the payload ID instead
         cFileName := aPayload[i,PL_ID]
      endif

      // remove any drive/path info from supplied filename
      cFileName := ParseFileName( cFileName )

      // erase any previous files with same name
      FErase( cUploadDir + cFileName )

      if aPayload[i,PL_DATA_T] == PL_DATA_FILE
         // the content was saved in a temp file - move the file to the upload folder
         if FSize(aPayload[i,PL_DATA]) == 0
            // don't bother saving zero length files
            // some browsers will send an empty from var even though no file was attached
            FErase( aPayload[i,PL_DATA] )
            loop
         endif
         FRename( aPayload[i,PL_DATA], cUploadDir + cFileName )
         if ! (lSuccess := FError() == 0)
            FErase( aPayload[i,PL_DATA] )
         endif
      else
         // save the data in memory to a file in the upload folder
         lSuccess := xbSaveToFile( aPayload[i,PL_DATA], cUploadDir + cFileName, .F.)
      endif

      if lSuccess
         oResponse:Content += cFileName + CRLF
      else
         oResponse:Content += "Error saving file: " + cFileName + CRLF
      endif
   next

   Return

//-----------------------------------------------------------------------------
STATIC FUNCTION CleanFileName( cName )
   Local i
   cName := AllTrim(cName)
   for i := len(cName) to 1 step -1
      if cName[i] == " "
         cName[i] := "_"
      elseif ! cName[i] $ "!$0123456789.ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz"
         cName[i] := "~"
      endif
   next
   Return cName

//-----------------------------------------------------------------------------
STATIC FUNCTION ParseFileName( cName )
   Local i
   if (i := RAt("\", cName)) > 0
      Return SubStr(cName, i+1)
   elseif (i := RAt(":", cName)) > 0
      Return SubStr(cName, i+1)
   elseif (i := RAt("/", cName)) > 0
      Return SubStr(cName, i+1)
   endif
   Return cName

//-----------------------------------------------------------------------------
// shows how to do client authentication
//
PROCEDURE WEB_WWWAuthenticate()
   Local oClient   := ThreadObject()
   Local oResponse := oClient:HTTPResponse
   Local cAuth     := oClient:HTTPRequest:Authorization
   Local cLogin

   if ! Empty(cAuth) .and. Upper(SubStr(cAuth,1,5)) == "BASIC"
      cLogin := xbBase64Decode(SubStr(cAuth,7))
   endif

   if ! cLogin == 'bugsbunny:whatsup'
      oResponse:StatusCode      := 401
      oResponse:WWWAuthenticate := 'Basic realm="Xb2.NET"'
      FlashError("Access denied!",,,.t.)
      Return
   endif

   FlashMessage("Excellent, you are now authenticated...")

   Return

//-----------------------------------------------------------------------------
PROCEDURE WEB_Login()
   Local lUpdate  := .t.
   Local cMessage := ""
   Local oClient  := ThreadObject()  // this is the xbHTTPThread object
   Local oServer  := oClient:Parent  // this is the xbHTTPServer object
   Local cUID     := oClient:GetVar("UID")
   Local cPWD     := oClient:GetVar("PWD")

   if Empty(cUID)
      // no user ID specified - redisplay login screen
      oClient:SendFile( oServer:RootDir + "\Login.htm" )
      Return
   endif

   USE password ALIAS pw INDEX password SHARED NEW

   cUID := Upper(Trim(cUID))

   if !(DbSeek(cUID, .f.) .and. Upper(Trim(pw->password)) == Upper(Trim(cPWD)))
      // invalid userid/password - go back to login screen
      oClient:SendFile( oServer:RootDir + "\Login.htm" )
      close all
      Return
   endif

   // we have a valid userid/password
   // the GetCargo/SetCargo methods will automatically open a session to keep track of this guy

   if oClient:GetCargo("UID") != NIL
      if oClient:GetCargo("UID") == cUID
         // same guy, don't update login count in password table
         lUpdate := .f.
      else
         // this client was previously logged in under a different userid
         cMessage := "<tr><td>You previously logged in as</td><td><b>" + oClient:GetCargo("UID") + "</b></td></tr>"
         // close the old session
         oClient:CloseSession()
      endif
   endif

   oClient:SetCargo("UID", cUID)
   oClient:SetCargo("SECURITY", pw->security)  // save the security level

   // attach the response string to the HTTP response object
   oClient:HTTPResponse:Content := UserInfoPage( oClient, oServer, cMessage )

   // update the login count if user was not already logged in
   if lUpdate .and. dbRLock()
      pw->nlogins   := pw->nlogins + 1
      pw->logindate := date()
      pw->logintime := time()
   endif

   // we don't need the password table anymore
   close all

   Return

//-----------------------------------------------------------------------------
PROCEDURE WEB_Logout()
   // close session
   ThreadObject():CloseSession()
   // display message for 2 sec. then take user to main screen
   FlashMessage("You are now logged out...", "/", 2)
   Return

//-----------------------------------------------------------------------------
PROCEDURE WEB_ChangePassword()
   Local cMessage
   Local oClient  := ThreadObject()           // this is the xbHTTPThread object
   Local oServer  := oClient:Parent           // this is the xbHTTPServer object
   Local cUID     := oClient:GetCargo("UID")  // and get the client's userid

   if Empty(cUID)
      NotLoggedIn( oClient )
      Return
   endif


   // OK, we have a valid user, let's save the data...

   USE password ALIAS pw INDEX password SHARED NEW

   if ! DbSeek(cUID, .f.)
      // for some reason that userid is no longer found - go back to login screen
      NotLoggedIn( oClient )
      close all
      Return
   endif

   if !dbRLock()
      cMessage := "<tr><td colspan=2><b><font color=red>ERROR: Unable to save changes</font></b></td></tr>"
      oClient:HTTPResponse:Content := UserInfoPage( oClient, oServer, cMessage )
      close all
      Return
   endif

   if !Empty(oClient:GetVar("PWD"))
      pw->password := oClient:GetVar("PWD")
   endif

   pw->text := oClient:GetVar("TXT")

   cMessage := "<tr><td colspan=2><b><font color=blue>Your changes have been saved</font></b></td></tr>"
   oClient:HTTPResponse:Content := UserInfoPage( oClient, oServer, cMessage )

   close all
   Return

//-----------------------------------------------------------------------------
PROCEDURE WEB_BrowseTable()
   Local cHtml
   Local oClient := ThreadObject()           // this is the xbHTTPThread object
   Local cUID    := oClient:GetCargo("UID")  // get the client's userid from session manager
   Local cTable  := oClient:GetVar("TBL")

   DEFAULT cTable TO "PASSWORD"

   // make sure user is logged in
   if Empty(cUID)
      NotLoggedIn( oClient )
      Return
   endif

   TEXT INTO cHtml TRIMMED
      <html><head>
      <title>Xb2.NET Browse Table</title>
      <meta name=viewport content="width=device-width, initial-scale=1">
      <link rel="stylesheet" type="text/css" href="Style.css"></head>
      <body bgcolor=black style=margin:0px;>
   ENDTEXT

   USE (cTable) INDEX (cTable) SHARED NEW

   oClient:HTTPResponse:Content := ;
           cHtml +;
           DrawTableGrid(oClient, cTable) +;
           "</body></html>"

   close all

   Return

//-----------------------------------------------------------------------------
STATIC FUNCTION UserInfoPage( oClient, oServer, cMessage )
   Local cHtml

   TEXT INTO cHtml TRIMMED
      <html><head>
      <title>Welcome to Xb2.NET</title>
      <meta name=viewport content="width=device-width, initial-scale=1">
      <link rel="stylesheet" type="text/css" href="Style.css"></head>
      <body bgcolor=black style=margin:0px;>
      <table width=80% align=center><tr><th colspan=2>
   ENDTEXT

   if pw->nlogins = 0
      cHtml += "Welcome " + Trim(pw->fname) +" "+ pw->lname +;
               "<br>this is your first login. Hope you have fun :)<br>&nbsp;</th></tr>" +;
               cMessage
   else
      cHtml += "Welcome back " + Trim(pw->fname) + "<br>&nbsp;</th></tr>" +;
               cMessage +;
               "<tr><td>Your login count</td><td>" + NTrim(pw->nlogins) + "</td></tr>" +;
               "<tr><td>Your last login</td><td>" + xbGMTDate(pw->logindate, pw->logintime) + "</td></tr>"
   endif

   cHtml += "<tr><td>Your security level</td><td>" + NTrim(pw->security) + "</td></tr>" +;
            "<tr><td>Your text message</td><td>" + pw->text + "</td></tr>" +;
            "<tr><td>Your unique session ID</td><td>" + oClient:GetSessionHandle() + "</td></tr>"

   cHtml += "<tr><th colspan=2><br>Server statistics:</th></tr>" +;
            "<tr><td>Server uptime</td><td>" + UpTime(oServer:UpTime()) + "</td></tr>" +;
            "<tr><td>Number of open sessions</td><td>" + NTrim(xbSession():ActiveSessions()) + "</td></tr>" +;
            "<tr><td>Number of open connections</td><td>" + NTrim(oServer:ActiveConnections()) + "</td></tr>" +;
            "<tr><td>Total connect count</td><td>" + NTrim(oServer:ConnectCount) + "</td></tr>"

   cHtml += "<tr><th colspan=2><br>Change Password & Text Message:</th></tr>" +;
            "<tr><td colspan=2><br>" +;
               "The data that you enter here will be saved in PASSWORD.DBF for UserID=" + Trim(pw->USERID)+ "<br><br>" +;
               "When you push the <b>Save</b> button, the data that you have entered will be submitted "+;
               "to the server as <i>NAME=VALUE</i> pairs and processed by the " +;
               "<A HREF=/View?'WebFunc.prg'>WEB_ChangePassword()</A> function. " +;
               "Along with the entered data, one additional hidden variable representing your unique session ID, " +;
               "will also be submitted. This session ID will be used by <i>WEB_ChangePassword</i> to:" +;
               "<ul>" +;
               "<li>verify that the client is logged in, and" +;
               "<li>position the record pointer on the correct record prior to writing changes." +;
               "</ul><br>" +;
            "<form name='pwd' method='post' action='ChangePassword'>" +;
            "<input type=hidden name='" + oServer:SessionTag + "' value='" + oClient:GetSessionHandle() + "'>"+;
            "<b>New Password:</b>&nbsp; <input type=password name='PWD' maxlength=10 size=16>&nbsp;&nbsp; " +;
            "<b>Text Message:</b>&nbsp; <input type=text name='TXT' maxlength=40 size=25>&nbsp;&nbsp; " +;
            "<input type=submit value='Save' style='font:bold;'>" +;
            "</form>" +;
            "</td></tr>"

   cHtml += "<tr><th style='text-align:center;'><br>" +;
            "<form name='brw' method='post' action='BrowseTable'>" +;
            "<input type=hidden name='" + oServer:SessionTag + "' value='" + oClient:GetSessionHandle() + "'>"+;
            "<input type=submit value='Browse Password File' style='font:bold;'>" +;
            "</form></th><th style='text-align:center;'><br>" +;
            "<form name='lout' method='post' action='Logout'>" +;
            "<input type=hidden name='" + oServer:SessionTag + "' value='" + oClient:GetSessionHandle() + "'>"+;
            "<input type=submit value='Logout' style='font:bold; color:navy;'>" +;
            "</form></th>" +;
            "</tr>"

   cHtml += "</table></body></html>"

   Return cHtml

//-----------------------------------------------------------------------------
STATIC FUNCTION UpTime( aTime )
   Local cTime := ""

   if aTime[UPTIME_DAY] != 0; cTime += NTrim(aTime[UPTIME_DAY]) + " d "; endif
   if aTime[UPTIME_HRS] != 0; cTime += NTrim(aTime[UPTIME_HRS]) + " h "; endif
   if aTime[UPTIME_MIN] != 0; cTime += NTrim(aTime[UPTIME_MIN]) + " m "; endif
   if aTime[UPTIME_SEC] != 0; cTime += NTrim(aTime[UPTIME_SEC]) + " s "; endif

   Return cTime

//-----------------------------------------------------------------------------
STATIC PROCEDURE NotLoggedIn( oClient )
   oClient:CloseSession()
   FlashError("Sorry, you need to log in first...", "Login.htm",,.t.)
   Return

//-----------------------------------------------------------------------------
STATIC FUNCTION DrawTableGrid( oClient, cTable, nMaxRows )
   Local i, cHtml, cPrev, cNext
   Local nRows    := 0
   Local nFields  := FCount()
   Local oURL     := xbURI():New()
   Local nDir     := iif(oClient:GetVar("DIR")==NIL, 1, Val(oClient:GetVar("DIR")))
   Local nFromRec := iif(oClient:GetVar("POS")==NIL, 0, Val(oClient:GetVar("POS")))

   DEFAULT nMaxRows TO 6

   oURL:Path("BrowseTable")
   oURL:SetVar(oClient:Parent:SessionTag, oClient:GetSessionHandle())
   oURL:SetVar("TBL", cTable)

   cHtml := "<table width=90% align=center><tr>"

   // draw column headings
   for i := 1 to nFields
      cHtml += "<th>" + FieldName(i) + "</th>"
   next

   cHtml += "</tr>"

   if nFromRec == 0
      dbGoTop()
   else
      dbGoTo(nFromRec)
      if nDir < 0
         // skip back nMaxRows
         dbSkip(-nMaxRows)
      endif
      // skip back to see if we're at the top
      // no sense of displaying "Prev Page" link if we're already at the top
      dbSkip(-1)
      if !Bof()
         dbSkip()
         oURL:SetVar("DIR", "-1")
         oURL:SetVar("POS", NTrim(RecNo()))
         cPrev := oURL:AsString()
      endif
   endif

   // draw the table contents
   while !Eof() .and. ++nRows <= nMaxRows
      cHtml += "<tr>"
      for i := 1 to nFields
         cHtml += "<td>" + Var2Char(FieldGet(i)) + "</td>"
      next
      cHtml += "</tr>"
      dbSkip()
   enddo

   if !Eof()
      // add page down link only if we're not at EOF
      oURL:SetVar("DIR", "1")
      oURL:SetVar("POS", NTrim(RecNo()))
      cNext := oURL:AsString()
   endif

   // add hyperlinks for page up/down and logout
   cHtml += "<tr><th colspan=" + NTrim(nFields) +" style='text-align:right;'><br>" +;
            iif(cPrev==Nil, "<font style=color:gray;>Prev Page</font>", "<a href=" + cPrev + ">Prev Page</a>") + ;
            " | " +;
            iif(cNext==Nil, "<font style=color:gray;>Next Page</font>", "<a href=" + cNext + ">Next Page</a>")

   oURL:Reset()
   oURL:Path("Logout")
   oURL:SetVar(oClient:Parent:SessionTag, oClient:GetSessionHandle())

   cHtml += " | <a href=" + oURL:AsString() + ">Logout</a>"+;
            "</th></tr>" +;
            "</table>"

   Return cHtml

//-----------------------------------------------------------------------------
// Ajax defines a web framework for dynamically pulling data from the server.
// Ajax makes it possible to update small sections of an HTML document
// with new server data without requiring a complete page refresh.
//
PROCEDURE WEB_AjaxTest( nAction )
   Static cHtml
   Local cResponse

   if cHtml == nil
   TEXT INTO cHtml TRIMMED
      <!doctype html>
      <html>
      <head>
      <title>Xb2.NET AjaxTest</title>
      <meta name=viewport content="width=device-width, initial-scale=1">
      <link rel='stylesheet' type='text/css' href='Style.css'></head>
      <style type='text/css'>
         h3 {color:#900; margin:0;}
      </style>
      <script src='prototype.js' type='text/javascript' ></script>
      <script language='javascript' type='text/javascript'>
      function ServerTime(){
        var ajax = new Ajax.Updater({success:'ajaxTime'}, 'AjaxTest(1)', {method:'get', parameters:''});
      }
      function ServerDate(){
        var ajax = new Ajax.Updater({success:'ajaxDate'}, 'AjaxTest(2)', {method:'post'});
      }
      </script>
      </head>

      <body onload="setInterval('ServerTime()', 1000);">

      <h1>Ajax Test</h1>
      This sample shows how to dynamically pull data from the server and update a small sec
      tion on this HTML document without requiring a complete page refresh.
      <br><br>
      <h3 id='ajaxTime'>Ajax data will load here...</h3>
      <h3 id='ajaxDate'></h3>
      <br>
      <input type="button" value="get server date" onclick="ServerDate()" />
      </body>
      </html>
   ENDTEXT
   endif

   do case
   case valtype(nAction) != "N"
      // if input param is not a number then we send initial page
      cResponse := cHtml
   case nAction == 1
      cResponse := "server time: " + time()
   otherwise
      cResponse := "server date: " + dtos(date())
   endcase

   ThreadObject():HTTPResponse:Content := cResponse

   Return

//-----------------------------------------------------------------------------
// Server push also known as "Comet" and "reverse Ajax"
// The Comet pattern is based on creating and maintaining long-lived HTTP
// connections that enable the server to push events or data to the client
// without waiting for the client to request it.
//
PROCEDURE WEB_CometTest( nAction )
   Static cHtml
   Local oResponse := ThreadObject():HTTPResponse


   if cHtml == nil
   TEXT INTO cHtml TRIMMED
      <!doctype html>
      <html>
        <head>
          <title>Xb2.NET CometTest</title>
          <meta name=viewport content="width=device-width, initial-scale=1">
          <link rel='stylesheet' type='text/css' href='style.css'></head>
          <script language='javascript'>
            function printTime(time) {
               document.getElementById('data').innerHTML = time;
            }
          </script>
        </head>

        <body>
          <h1>Comet Test</h1>
          This sample shows how to push events or data to the client without waiting for the client to request it.
          <br><br><br>
          <code id='data'>This is where server data is pushed</code>
          <hr>
          <h2>
          <a href='CometTest(1)?' target='iframe'/>start clock</a>
          </h2>
          <iframe id='iframe' name='iframe' width=0 height=0 border=0></iframe>
        </body>
      </html>
   ENDTEXT
   endif

   // if input param is not a number then we send initial page
   if valtype(nAction) != "N"
      oResponse:Content := cHtml
      Return
   endif

   // Use chunked transfer encoding to periodically push small bits of data
   // (the current server time) to client. We continue sending new data until
   // connection is terminated (browser closed, user navigates off page, etc..)
   //
   // Each chunk contains some javascript that is automatically executed by
   // the browser as it's received. In the Xbase++ world this is analogous to
   // sending character strings containing macros or codeblocks from one app
   // to be executed in another.

   oResponse:TransferEncoding('chunked')
   // send message header
   oResponse:Send()

   while oResponse:SendChunk("<script>parent.printTime('" + time() + "');</script>") != SOCKET_ERROR
      // pause for one second before sending the next chunk
      ? ThreadId(), "CometTest sending chunk"
      sleep(100)
   end
   ? ThreadID(), "CometTest stopped"

   Return

